from requests import Response
from otscrape.core.base.attribute import Attribute


class RequestText(Attribute):
    def __init__(self, target=None, *, bytes_result=False, encoding=None, project=True, replace_error=None):
        super().__init__(target=target, project=project, replace_error=replace_error)

        self.bytes_result = bytes_result
        self.encoding = encoding

    def extract(self, page):
        x = page[self.target]
        assert isinstance(x, Response)

        if self.bytes_result:
            return x.content
        elif self.encoding:
            return x.content.decode(self.encoding)

        return x.text


class RequestStatusCode(Attribute):

    def extract(self, page):
        target = self.target
        assert isinstance(page[target], Response)

        return page[target].status_code


class RequestJSON(Attribute):

    def extract(self, page):
        target = self.target
        assert isinstance(page[target], Response)

        return page[target].json()



from otscrape.core.base.mixins import NoFailMixin


class AttributeBase:
    def __init__(self, target=None, project=True):
        self.target = target or 'raw'
        self.do_project = project

    def __call__(self, page):
        try:
            return self.extract(page)
        except Exception as e:
            return self._on_extract_error(e)

    def _on_extract_error(self, exception):
        return self.on_extract_error(exception)

    def on_extract_error(self, exception):
        raise exception

    def extract(self, page):
        raise NotImplementedError()


class Attribute(NoFailMixin, AttributeBase):
    def __init__(self, target=None, project=True, replace_error=None):
        super().__init__(target=target, project=project)
        self.replace_error = replace_error

    @property
    def _return_value_when_fail(self):
        return self.replace_error

    def on_extract_error(self, *args, **kwargs):
        message = f'An error occurred while applying {self.__class__.__name__}() to attribute "{self.target}".'
        return self.on_error(*args, message=message, **kwargs)


def attribute(func=None, *, project=True, replace_error=None):
    if func:
        x = Attribute(project=project, replace_error=replace_error)
        x.extract = func
        return x

    def func_(f):
        x_ = Attribute(project=project, replace_error=replace_error)
        x_.extract = f
        return x_
    return func_



from otscrape.core.base.loader import Loader


class DummyLoader(Loader):
    def __init__(self, data=None):
        self.data = data

    def __call__(self, data=None):
        data = data or self.data
        return data



from copy import deepcopy
import requests

from otscrape.core.base.mixins import NoFailMixin
from otscrape.core.base.loader import Loader


class RequestLoaderBase(Loader):
    def __init__(self, method=None, accept_status_codes=(200,), **kwargs):
        super().__init__()

        self.method = method or 'GET'
        self.kwargs = kwargs or {}
        self.accept_status_codes = accept_status_codes

    def __deepcopy__(self, memo):
        cls = self.__class__
        result = cls.__new__(cls)
        memo[id(self)] = result
        for k, v in self.__dict__.items():
            setattr(result, k, deepcopy(v, memo))
        return result

    def post(self, **kwargs):
        result = deepcopy(self)
        result.method = 'POST'
        for k, v in kwargs.items():
            setattr(result, k, v)
        return result

    def get(self, **kwargs):
        result = deepcopy(self)
        result.method = 'GET'
        for k, v in kwargs.items():
            setattr(result, k, v)
        return result

    def make_request(self, url, **kwargs):
        kwargs_update = dict(self.kwargs)
        kwargs_update.update(kwargs)
        kwargs_update['url'] = url

        resp = requests.request(self.method, **kwargs_update)
        if resp.status_code in self.accept_status_codes:
            return resp
        resp.raise_for_status()

    def __call__(self, url, **kwargs):
        self.on_requesting()

        try:
            return self.make_request(url, **kwargs)
        except Exception as e:
            return self.on_request_error(e)
        finally:
            self.on_requested()

    def _on_requesting(self):
        return self.on_requesting()

    def _on_requested(self):
        return self.on_requested()

    def _on_request_error(self, exception):
        return self.on_request_error(exception)

    def on_requesting(self):
        pass

    def on_requested(self):
        pass

    def on_request_error(self, exception):
        raise exception


class SimpleRequestLoader(NoFailMixin, RequestLoaderBase):
    def __init__(self, method=None, accept_status_codes=(200,), replace_error=None, **kwargs):
        super().__init__(method=method, accept_status_codes=accept_status_codes, **kwargs)
        self.replace_error = replace_error

    @property
    def _return_value_when_fail(self):
        return self.replace_error

    def on_request_error(self, *args, **kwargs):
        return self.on_error(*args, **kwargs)



try:
    from functools import cached_property
except ImportError:
    from .futures import cached_property

from .attribute import AttributeBase


class DataModelMeta(type):
    def __new__(mcs, name, bases, dct):
        attr_ids = {}
        attrs = {}
        proj = set()

        # add attributes
        for name, obj in dct.items():
            if isinstance(obj, AttributeBase):
                attr_ids[id(obj)] = name
                attrs[name] = cached_property(obj)
                if getattr(obj, 'do_project', False):
                    proj.add(name)

        # add attributes from parents
        for base in bases:
            if not hasattr(base, '_attributes'):
                continue
            for name in getattr(base, '_attributes'):
                attrs[name] = getattr(base, name)
        dct_new = dict(dct)
        dct_new.update(attrs)

        dct_new['_attributes'] = set(attrs.keys())
        dct_new['_attribute_ids'] = attr_ids
        dct_new['_project_attrs'] = proj

        instance = super().__new__(mcs, name, bases, dct_new)
        return instance


class DataModel(metaclass=DataModelMeta):
    _attributes = []
    _project_attrs = set()
    _attribute_ids = {}

    def __init__(self):
        pass

    def get_data(self):
        result = {}
        for key in self._attributes:
            if key in self._project_attrs:
                result[key] = getattr(self, key)
        return result

    def __getitem__(self, key):
        if isinstance(key, AttributeBase):
            id_ = id(key)
            if id_ in self._attribute_ids:
                key = self._attribute_ids[id_]
            else:
                raise ValueError(f'attribute {key} doesn\'t belong to this page')

        if isinstance(key, str):
            if key not in self._attributes:
                raise ValueError(f'attribute {key} not found')

            return getattr(self, key)
        else:
            raise TypeError(f'attribute {key} should be of type str or AttributeBase')



from requests import Response

from otscrape.core.base.attribute import Attribute

from .request import RequestJSON
from .text.json import JSONDict
from .dict import get_item_from_path


parsers = [
    (Response, RequestJSON),
    ((str, bytes), JSONDict),
]


class JSON(Attribute):
    def __init__(self, target=None, path='/', *, project=True, replace_error=None, **kwargs):
        super().__init__(target=target, project=project, replace_error=replace_error)

        self.path = path
        self._kwargs = kwargs

        self._parsers = [(c, p(target, ** self._kwargs)) for c, p in parsers]

    def extract(self, page):
        target = self.target
        cache_name = f'{target}#JSON'

        if cache_name in page.cached:
            y = page.cached[cache_name]
        else:
            x = page[target]

            for c, p in self._parsers:
                if isinstance(x, c):
                    y = p.extract(page)
                    break
            else:
                raise TypeError(f'Unexpected type {x.__class__.__name__} for a JSON attribute.')

            page.cached[cache_name] = y

        item = get_item_from_path(y, self.path)
        return item



from requests import Response

from otscrape.core.base.attribute import Attribute
from .text.etree import parse
from .request import RequestText


class XPath(Attribute):
    def __init__(self, target=None, xpath='.', *, only_first=False, encoding=None, project=True, replace_error=None):
        super().__init__(target=target, project=project, replace_error=replace_error)

        self.xpath = xpath
        self.only_first = only_first
        self._req_parser = RequestText(target, encoding=encoding)

    def extract(self, page):
        target = self.target
        x = page[target]

        if isinstance(x, Response):
            x = self._req_parser.extract(page)
        elif not isinstance(x, (str, bytes)):
            raise TypeError(f'Unexpected type {x.__class__.__name__} for a XPath attribute.')

        cache_name = f'{target}#ETree'

        if cache_name in page.cached:
            tree = page.cached[cache_name]
        else:
            tree = parse(x)
            page.cached[cache_name] = tree

        result = tree.xpath(self.xpath)
        if self.only_first:
            result = result[0]

        return result



from typing import Callable

from otscrape.core.base.data_model import DataModel
from otscrape.core.base.attribute import attribute


class Page(DataModel):
    def __init__(self, **kwargs):
        super().__init__()

        self._loader_kwargs = kwargs
        self.cached = {}

    @property
    def name(self) -> str:
        raise NotImplementedError()

    @property
    def loader(self) -> Callable:
        raise NotImplementedError()

    @attribute(project=False)
    def raw(self):
        return self.fetch()

    def fetch(self):
        return self.loader(**self._loader_kwargs)



from otscrape.core.base.attribute import attribute, Attribute
from otscrape.core.attribute import RequestText, RequestStatusCode, JSON, DictPath, XPath, ZipDict

from otscrape.core.io import SimpleRequestLoader, DummyLoader

from otscrape.core.base.page import Page



import traceback


class NoFailMixin:
    _return_value_when_fail = None

    def on_error(self, exception, message=None):
        try:
            if hasattr(super(), 'on_error'):
                super().on_error(exception)
        except Exception as e:
            return self.__handle(e, message)

        return self.__handle(exception, message)

    def __handle(self, exception, message=None):
        traceback.print_exception(type(exception), exception, exception.__traceback__)
        if message:
            exception = exception.__class__(message).with_traceback(exception.__traceback__)
        traceback.print_exception(type(exception), exception, None)

        value = self._return_value_when_fail
        if callable(value):
            value = value()
        return value



from typing import Iterator, Iterable

from otscrape.core.base.attribute import AttributeBase, Attribute


class ZipDict(Attribute):
    def __init__(self, structure=None, *, project=True, replace_error=None):
        super().__init__(target=None, project=project, replace_error=replace_error)

        assert structure and all(isinstance(x, (str, AttributeBase)) for x in structure.values())
        self.structure = structure

    def extract(self, page):
        iter_dict = {}
        for key, value in self.structure.items():
            if isinstance(page[value], Iterable):
                iter_dict[key] = iter(page[value])
            else:
                iter_dict[key] = page[value]

        result = []
        try:
            while True:
                row = {}
                for key, it in iter_dict.items():
                    if isinstance(it, Iterator):
                        row[key] = next(it)
                    else:
                        row[key] = it
                result.append(row)
        except StopIteration:
            pass

        return result



import re

from otscrape.core.base.attribute import Attribute


key_pattern = re.compile(r'^[a-zA-Z_]+$')
key_index_pattern = re.compile(r'^[a-zA-Z_]+\[\d+\]$')


def get_item_from_path(d, path):
    assert path.startswith('/') and '//' not in path

    for key in (s for s in path.split('/') if s):
        if key_pattern.match(key):
            assert key in d
            d = d[key]
        elif key_index_pattern.match(key):
            name, index = key.split('[')
            index = int(index.strip(']'))

            assert name in d
            d = d[name]

            assert hasattr(d, '__getitem__')
            d = d[index]

    return d


class DictPath(Attribute):
    def __init__(self, target=None, path='/', *, project=True, replace_error=None):
        super().__init__(target=target, project=project, replace_error=replace_error)

        self.path = path

    def extract(self, page):
        target = self.target
        item = get_item_from_path(page[target], self.path)
        return item



from lxml import etree
from otscrape.core.base.attribute import Attribute


def parse(text):
    tree = etree.fromstring(text, etree.HTMLParser())
    return tree


class ETree(Attribute):
    def extract(self, page):
        x = page[self.target]
        assert isinstance(x, (str, bytes))

        return parse(x)




import setuptools
from setuptools import setup


with open('requirements.txt') as f:
    requirements = f.read().splitlines()

setup(
    name='one-two-scrape',
    version='0.0.1',
    packages=setuptools.find_packages(),
    url='',
    license='Apache 2.0',
    author='SSripilaipong',
    author_email='SHSnail@mail.com',
    description='',
    python_requires='>=3.6',
    install_requires=requirements,
)



import json

from otscrape.core.base.attribute import Attribute


class JSONDict(Attribute):
    def __init__(self, target=None, *, encoding=None, project=True, replace_error=None, **kwargs):
        super().__init__(target=target, project=project, replace_error=replace_error)

        self.encoding = encoding
        self.kwargs = kwargs

    def extract(self, page):
        target = self.target
        assert isinstance(page[target], (str, bytes))

        return json.loads(page[target], encoding=self.encoding, **self.kwargs)
